#!/usr/bin/env python

"""
setup.py file for feffpathwrapper (via SWIG)
"""

from distutils.core import setup, Extension
import sys, os
from os.path import join

import larch
#installdir = larch.site_config.larchdir
installdir = larch.larchlib.larchdir
dlldir =  larch.larchlib.get_dlldir()

if os.name == 'nt':
    feffpath_module = Extension('_feffpathwrapper',
                                sources=['feffpath_wrap.c'],
                                include_dirs = ['C:\\Program Files\\mingw-w64\\x86_64-4.9.2-win32-seh-rt_v3-rev1\\mingw64\\lib\\gcc\\x86_64-w64-mingw32\\4.9.2\\include',],
                                library_dirs = [join('..','..','src','GENFMT')],
                                libraries=['feffpath'],
                            )
else:
    feffpath_module = Extension('_feffpathwrapper',
                                sources=['feffpath_wrap.c'],
                                library_dirs = [join('..','..','src','GENFMT')],
                                libraries=['feffpath'],
                            )

setup (name         = 'scatteringpath',
       version      = '0.1',
       author       = "Bruce Ravel",
       author_email = 'bravel@bnl.gov',
       url          = 'https://github.com/xraypy/feff85exafs',
       download_url = 'https://github.com/xraypy/feff85exafs',
       description  = """A Larch wrapper around the feffpath library""",
       ext_modules  = [feffpath_module],
       data_files   = [(join(installdir,'plugins','xafs'), ["scatteringpath.py",]),
                       (join(installdir,'modules'),        ["feffpathwrapper.py",]),
                       (join(installdir,'dlls',dlldir),    ["_feffpathwrapper.so",])],
       )



from scatteringpath import scatpath

a=scatpath()
epsilon=1e-4

def test_phpad():
    yield check_phpad

def test_degen():
    yield check_degen, 0
    yield check_degen, 48

def test_index():
    yield check_index, 0
    yield check_index, 1

def test_iorder():
    yield check_iorder, 0
    yield check_iorder, 1


def test_nnnn():
    yield check_nnnn, True
    yield check_nnnn, False

def test_verbose():
    yield check_verbose, True
    yield check_verbose, False


def test_evec():
    yield check_evec

def test_xivec():
    yield check_xivec

def test_atom():
    a.clear()
    yield check_index, 0
    a.degen = 48
    a.index = 4
    a.phpad = "../fortran/phase.pad"
    a.nnnn  = False
    a.atom(0, 0, -3.61, 1)
    yield check_atom, 2
    a.atom(-1.805, 0, -1.805, 1)
    yield check_atom, 3
    a.make()
    yield check_ri, 'ri'
    yield check_beta, 'beta'
    yield check_param, 'edge'
    yield check_param, 'gam_ch'
    yield check_param, 'kf'
    yield check_param, 'mu'
    yield check_param, 'rnorman'
    yield check_param, 'rs_int'
    yield check_param, 'vint'
    yield check_param, 'exch'

def test_atom_errors():
    a.clear();
    a.atom(0, 0, -3.61, -1)     # ipot negative
    yield check_error, 1

    a.clear();
    a.atom(0, 0, -3.61, 9)      # ipot too big
    yield check_error, 2

    a.clear();
    a.atom(0, 0, -3.61, 1)
    a.atom(0, 0, -3.71, 1)
    yield check_error, 4        # atoms too close together

    a.clear();
    a.atom(0, 0, 0, 1)
    a.atom(0, 0, -3.61, 1)
    a.make()
    yield check_error, 1        # first atom absorber

    a.clear();
    a.atom(0, 0, -3.61, 1)
    a.atom(0, 0, 0, 1)
    a.make()
    yield check_error, 2        # last atom absorber

    a.clear();
    try:
        a.degen=-12
    except ValueError:
        yield check_pythonerror, 'degen'        # degeneracy negative
    else:
        yield check_wrong_pythonerror, 'degen'

    try:
        a.index=40000
    except ValueError:
        yield check_pythonerror, 'index'        # bad index
    else:
        yield check_wrong_pythonerror, 'index'

    try:
        a.elpty=-0.5
    except ValueError:
        yield check_pythonerror, 'elpty'        # bad elpty
    else:
        yield check_wrong_pythonerror, 'elpty'

    try:
        a.iorder=-1
    except ValueError:
        yield check_pythonerror, 'iorder'        # bad elpty
    else:
        yield check_wrong_pythonerror, 'iorder'

    try:
        a.phpad='foo.bar'
    except ValueError:
        yield check_pythonerror, 'phpad'        # bad phpad
    else:
        yield check_wrong_pythonerror, 'phpad'


################################################################################


def check_phpad():
    pb = "../fortran/phase.pad";
    a.phpad = pb
    assert a.wrapper.phpad == pb

def check_degen(degen):
    if degen:
        a.degen = degen
        assert a.wrapper.degen == degen
    else:
        assert a.degen == 1

def check_index(ind):
    if ind:
        a.index = ind
        assert a.wrapper.index == ind
    else:
        assert a.index == 9999

def check_iorder(iorder):
    if iorder:
        a.iorder = iorder
        assert a.wrapper.iorder == iorder
    else:
        assert a.iorder == 2

def check_nnnn(nnnn):
    a.nnnn = nnnn
    assert a.wrapper.nnnn == nnnn

def check_verbose(verbose):
    a.verbose = verbose
    assert a.wrapper.verbose == verbose

def check_evec():
    a.evec = [0,0,1]
    assert a.evec[0] == 0 and a.evec[1] == 0 and a.evec[2] == 1

def check_xivec():
    a.evec = [1,1,0]
    assert a.evec[0] == 1 and a.evec[1] == 1 and a.evec[2] == 0

def check_atom(nleg):
    assert a.nleg == nleg

def check_ri(which):
    assert abs(a.ri[0] - 3.61) < epsilon and abs(a.ri[1] - 2.55266) < epsilon and abs(a.ri[2] - 2.55266) < epsilon

def check_beta(which):
    assert abs(a.beta[0] - 135) < epsilon and abs(a.beta[1] - 90) < epsilon and abs(a.beta[2] - 135) < epsilon

def check_error(err):
    assert a.errorcode == err

def check_pythonerror(which):
    assert True
def check_wrong_pythonerror(which):
    assert False

def check_param(which):
    if (which == 'edge'):
        assert abs(a.edge + 3.82035) < epsilon
    elif (which == 'gam_ch'):
        assert abs(a.gam_ch - 1.72919) < epsilon
    elif (which == 'kf'):
        assert abs(a.kf - 1.824) < epsilon
    elif (which == 'mu'):
        assert abs(a.mu + 3.82035) < epsilon
    elif (which == 'rnorman'):
        assert abs(a.rnorman - 2.63173) < epsilon
    elif (which == 'rs_int'):
        assert abs(a.rs_int - 1.98947) < epsilon
    elif (which == 'vint'):
        assert abs(a.vint + 16.48140) < epsilon
    elif (which == 'exch'):
        assert a.exch == 'H-L exch'
    else:
        assert False;






from os.path import realpath, exists, join

from larch import (Group, param_group, Parameter, isParameter, param_value, use_plugin_path, isNamedClass)
use_plugin_path('io')
from xdi import read_xdi
use_plugin_path('xafs')
from feffit import feffit_dataset, feffit_transform, feffit, feffit_report
from feffdat import feffpath
use_plugin_path('wx')
from plotter import (_newplot, _plot)


def do_fit(self, which):

    if which == 'testrun':
        folder = self.testrun
    else:
        folder = self.baseline

    data = read_xdi(join(self.path, 'NiO.chik'), _larch=self._larch)
    if hasattr(data, 'wavenumber'):
        data.k = data.wavenumber

    gds = param_group(amp    = Parameter(1,      vary=True),
                      enot   = Parameter(0.01,   vary=True),
                      alpha  = Parameter(0.0001, vary=True),
                      sso    = Parameter(0.003,  vary=True),
                      ssni   = Parameter(0.003,  vary=True),
                      sso2   = Parameter(0.003,  vary=True),
                      #sso3   = Parameter(0.003,  vary=True),
                      ssni2  = Parameter(0.003,  vary=True),
                      #ssni3  = Parameter(0.003,  vary=True),
                      #ssni4  = Parameter(0.003,  vary=True),
                      _larch=self._larch  )

    paths = list()
    paths.append(feffpath(realpath(join(folder, "feff0001.dat")), # 1st shell O SS
                          s02    = 'amp',
                          e0     = 'enot',
                          sigma2 = 'sso',
                          deltar = 'alpha*reff', _larch=self._larch))
    paths.append(feffpath(realpath(join(folder, "feff0002.dat")), # 2nd shell Ni SS
                          s02    = 'amp',
                          e0     = 'enot',
                          sigma2 = 'ssni',
                          deltar = 'alpha*reff', _larch=self._larch))
    paths.append(feffpath(realpath(join(folder, "feff0003.dat")), # O-O triangle
                          s02    = 'amp',
                          e0     = 'enot',
                          sigma2 = '1.5*sso',
                          deltar = 'alpha*reff', _larch=self._larch))
    paths.append(feffpath(realpath(join(folder, "feff0004.dat")), # O-Ni triangle
                          s02    = 'amp',
                          e0     = 'enot',
                          sigma2 = 'sso+ssni',
                          deltar = 'alpha*reff', _larch=self._larch))
    paths.append(feffpath(realpath(join(folder, "feff0005.dat")), # 3rd shell O SS
                          s02    = 'amp',
                          e0     = 'enot',
                          sigma2 = 'sso2',
                          deltar = 'alpha*reff', _larch=self._larch))
    paths.append(feffpath(realpath(join(folder, "feff0006.dat")), # 4th shell Ni SS
                          s02    = 'amp',
                          e0     = 'enot',
                          sigma2 = 'ssni2',
                          deltar = 'alpha*reff', _larch=self._larch))
    paths.append(feffpath(realpath(join(folder, "feff0007.dat")), # O-O non-forward linear
                          s02    = 'amp',
                          e0     = 'enot',
                          sigma2 = 'sso*2',
                          deltar = 'alpha*reff', _larch=self._larch))
    paths.append(feffpath(realpath(join(folder, "feff0008.dat")), # O-Ni forward scattering
                          s02    = 'amp',
                          e0     = 'enot',
                          sigma2 = 'ssni2',
                          deltar = 'alpha*reff', _larch=self._larch))
    paths.append(feffpath(realpath(join(folder, "feff0009.dat")), # O-O forward through absorber
                          s02    = 'amp',
                          e0     = 'enot',
                          sigma2 = 'sso*2',
                          deltar = 'alpha*reff', _larch=self._larch))
    paths.append(feffpath(realpath(join(folder, "feff0010.dat")), # O-Ni-O double forward
                          s02    = 'amp',
                          e0     = 'enot',
                          sigma2 = 'ssni2',
                          deltar = 'alpha*reff', _larch=self._larch))


    trans = feffit_transform(kmin=3, kmax=15.938, kw=(2,1,3), dk=1, window='hanning', rmin=1.5, rmax=4.2, _larch=self._larch)
    dset  = feffit_dataset(data=data, pathlist=paths, transform=trans, _larch=self._larch)
    fit   = feffit(gds, dset, _larch=self._larch)

    if self.doplot:
        offset = 0.6*max(dset.data.chir_mag)
        _newplot(dset.data.r,  dset.data.chir_mag+offset, xmax=8, win=2,
              xlabel=r'$R \rm\,(\AA)$', label='data',
              ylabel=r'$|\chi(R)| \rm\,(\AA^{-3})$',
              title='Fit to '+self.folder, show_legend=True, _larch=self._larch)
        _plot(dset.model.r, dset.model.chir_mag+offset, label='fit', win=2, _larch=self._larch)
        _plot(dset.data.r,  dset.data.chir_re, label='data', win=2, _larch=self._larch)
        _plot(dset.model.r, dset.model.chir_re, label='fit', win=2, _larch=self._larch)
    #end if

    if self.verbose:
        print feffit_report(fit, _larch=self._larch)
    #end if

    return fit
#end def




## feff85exafs unit testing system using larch
## see HEADERS/license.h for feff's license information

from   os        import makedirs, chdir, getcwd, unlink, listdir
from   os.path   import realpath, isdir, isfile, join
from   shutil    import rmtree, copy
import sys, subprocess, glob, pystache, json, re
from   termcolor import colored
import numpy     as np
import importlib
from   distutils.spawn import find_executable

from larch import (Group, Parameter, isParameter, param_value,
                   isNamedClass, Interpreter)
from larch_plugins.xafs.feffdat import feffpath
from larch_plugins.xafs.feffrunner import feffrunner
from larch_plugins.wx import (_newplot, _plot)

WRAPPER_AVAILABLE = False
try:
    from larch_plugins.xafs.feff8lpath import Feff8L_XAFSPath
    WRAPPER_AVAILABLE = True
except:
    pass


WARN_COLOR = 'yellow'
WARN_COLOR = 'blue'

def print_error(text):
    print(colored(text, 'magenta', attrs=['bold']))

def print_warn(text):
    print(colored(text, WARN_COLOR, attrs=['bold']))

def print_good(text):
    print(colored(text, 'green', attrs=['bold']))

def test_text(text, cond):
    color = 'green' if cond else 'magenta'
    return colored(text, color, attrs=['bold'])

class Feff85exafsUnitTestGroup(Group):
    """
    A group for performing unit tests on feff85exafs.

    Activity methods:
       run            : run the test feff calculation, no return value
       fit            : run fits using the baselline and new Feff caclautions
       clean          : remove the testrun folder

    Testing methods:
       available      : returns True is a path index has a corresponding feffNNNN.dat file
       compare        : make a comparison of columns in feffNNNN.dat, returns True if no difference
       feffterms      : perform a test on values in the header of feffNNNN.dat, ret. True if no difference
       radii          : fetch a list of muffin tin or norman radii for the unique potentials
       s02            : fetch the calculated value of s02 from the testrun or the baseline
       print_geometry : write a description of scattering path to the screen

    Attributes:
       doplot    :  boolean, True = make plots in compare method
       doscf     :  boolean, True = use self-consistency when running feff for unit tests
       verbose   :  boolean, True = write screen messages when running feff and performing tests
       feffran   :  boolean, True = feff has been run and testrun folder holds the output
       folder    :  string,  name of folder containing test materials
       testrun   :  string,  name of folder containing output of feff test run
       fefflog   :  string,  name of log file from feff run
       baseline  :  string,  name of folder containing the baseline feff calculation
       paths     :  list,    list of feffNNNN.dat files in testrun folder
       bpaths    :  list,    list of feffNNNN.dat files from baseline calculation
       path      :  string,  fully resolved path to folder
       repotop   :  string,  fully resolved path to top of feff85exafs repository
       json      :  json string, used to configure the test feff run
       rfactor   :  float,   R-factor computed from feffNNNN.dat column comparison, set by compare()
       rfactor_2 :  float,   second R-factor, used when compare called with part='feff', set by compare()
       epsilon   :  float,   value for comparing columns from feffNNNN.dat with the baseline and other things
       count     \\
       datacount  > integers, used to count number of tests run
       feffcount /
    """

    def __init__(self, folder=None, _larch=None, **kws):
        kwargs = dict(name='Feff85exafs unit test: %s' % folder)
        kwargs.update(kws)
        Group.__init__(self,  **kwargs)
        self._larch     = Interpreter()
        self.doplot     = True
        self.doscf      = False # True = use self-consistency
        self.verbose    = True  # True = print Feff's screen messages and other screen messages
        self.feffran    = False # True = Feff calculation has been run
        self.count      = 0
        self.feffcount  = 0
        self.datacount  = 0
        self.failed     = list()
        if folder[-1] == '/': folder = folder[:-1] # strip trailing /
        self.folder     = folder

        if not isdir(folder):
            folder = join('tests', folder)
        if not isdir(folder):
            print_error(folder + " isn't one of the available tests")
            return None

        self.path       = realpath(folder)
        self.testrun    = realpath(join(self.path, 'testrun'))
        self.fefflog    = realpath(join(self.path, 'testrun', 'feff8l.log'))
        self.__testpaths()
        self.repotop    = getcwd()
        if not self.repotop.endswith('feff85exafs'):  self.repotop = realpath(join('..'))
        # the f85e shell script emulates the behavior of the monolithic Feff application
        self.eps5       = 0.00001
        self.eps4       = 0.0001
        self.eps3       = 0.001
        self.epsilon    = self.eps4
        self.epsfit     = self.eps3
        self.epserr     = 5.0 * self.epsfit
        self.firstshell = False
        self.fittest    = None
        if WRAPPER_AVAILABLE:
            self.sp = Feff8L_XAFSPath(_larch=self._larch)

    def __repr__(self):
        if not isdir(self.folder):
            return '<Feff85exafs Unit Test Group (empty)>'
        if self.folder is not None:
            return '<Feff85exafs Unit Test Group: %s>' % self.folder
        return '<Feff85exafs Unit Test Group (empty)>'


    @property
    def baseline(self):
        scf = 'withSCF' if self.doscf else 'noSCF'
        return realpath(join(self.path, 'baseline', scf))

    @property
    def bpaths(self):
        """
        Gather a list of feffNNNN.dat files from the baseline calculation

        """
        p   = list()
        owd = getcwd()
        try:
            chdir(self.baseline)
            feffoutput = glob.glob("*")
            for f in sorted(feffoutput):
                tosave = re.compile("feff\d+\.dat")
                if tosave.match(f):
                    p.append(f)
        finally:
            chdir(owd)
        return p

    def run(self):
        """
        Make a feff.inp from the mustache template, then run feff 8.5 in
        the testrun folder

        """
        if not isdir(self.path):
            print_error(self.folder + " is not one of the available tests")
            return False

        ## clean up an earlier run
        if isdir(self.testrun): rmtree(self.testrun)
        makedirs(self.testrun)

        ## mustache to feff.inp
        scf = 'without SCF'
        self.json = json.load(open(join(self.path, self.folder + '.json')))
        self.json['doscf']='* '
        if self.doscf:
            scf = 'with SCF'
            self.json['doscf']=''
        renderer = pystache.Renderer()
        with open(join(self.testrun,'feff.inp'), 'w') as inp:
            inp.write(renderer.render_path( join(self.path, self.folder + '.mustache'), # material/material.mustache
                                            self.json ))                                # material/material.json

        ## run feff with feffrunner
        owd = getcwd()
        try:
            chdir(self.testrun)
            if isfile(self.fefflog):
                unlink(self.fefflog)

            self.feffrunner=feffrunner(feffinp=join(self.testrun,'feff.inp'), verbose=self.verbose, repo=self.repotop, _larch=self._larch)
            self.feffrunner.run()
            for f in glob.glob("*"):
                if f.startswith('log'):
                    unlink(f)
            if self.verbose:
                print_warn("\nRan Feff85EXAFS on %s (%s)" % (self.folder, scf))
            self.__testpaths()
        finally:
            chdir(owd)
        self.feffran = True

    def run_opconsat(self):
        if not self.feffran:
            print_error("You need to run the rest of the feff calculation first.")
            return False
        ## run feff with feffrunner
        owd = getcwd()
        try:
            chdir(self.testrun)
            self.feffrunner=feffrunner(feffinp=join(self.testrun,'feff.inp'), verbose=self.verbose, repo=self.repotop, _larch=self._larch)
            copy(join(self.testrun, "..", "opconsat", "baseline", "exc.inp"), self.testrun)
            self.feffrunner.run('opconsat')
        finally:
            chdir(owd)
        if not isfile(join(self.testrun, "exc.dat")):
            print_error("Failed to run opconsat.")
            return False
        return True

    def __testpaths(self):
        """
        Gather a list of feffNNNN.dat files from the testrun

        """
        self.paths = list()
        if isdir(self.testrun):
            owd = getcwd()
            try:
                chdir(self.testrun)
                feffoutput = glob.glob("*")
                for f in sorted(feffoutput):
                    tosave = re.compile("feff\d+\.dat")
                    if tosave.match(f):
                        self.paths.append(f)
                self.feffran = True
            finally:
                chdir(owd)


    def available(self, which=0):
        """
        Test whether a path index was included in the test calculation.

           larch> print group.available(1)
           True

           larch> print group.available(112)
           False

        If the argument is 0 or negative, print a list of all available paths files and return None

        """
        if which>0:
            nnnn = "feff%4.4d.dat" % which
            if self.paths.count(nnnn) and self.bpaths.count(nnnn):
                return True
            else:
                return False
        else:
            print "The following paths are available:"
            for f in self.paths:
                print "\t"+f
            return None


    def __snarf_geometry(self, index=1):
        pathsdat = self.testrun+'/paths.dat'
        try:
            lines = open(pathsdat, 'r').readlines()
        except:
            print( 'Error reading file %s ' % pathsdat)
            return

        degmatch = re.compile("degeneracy")
        mode = 'searching'
        count = 1
        for line in lines:
            line = line.replace(',', ' ').strip()
            words  = line.split()
            if len(words) < 5:
                continue
            elif mode == 'searching':
                #print '> %s %s' % (line, str(degmatch.match(line)))
                if words[5] == 'degeneracy' and line.startswith(str(index)):
                    self.sp.index = index
                    nlegs = int(words[1])
                    self.sp.deg = float(words[2])
                    mode = 'parsing'
                    continue
            elif mode == 'parsing':
                #print '| %s %d %d' % (line, count, nlegs)
                if words[5] == 'degeneracy':
                    break
                elif count == nlegs:
                    break
                elif line.startswith('x'):
                    continue
                else:
                    words = line.split()
                    self.sp.add_scatterer(x=float(words[0]),
                                          y=float(words[1]),
                                          z=float(words[2]),
                                          ipot=int(words[3]))
                    count+=1

    def compare(self, nnnn=1, part='feff', use_wrapper=False, _larch=None):
        """
        Compare a feffNNNN.dat file from the testrun with the same file
        from the baseline calculation

            group.compare(N, part, use_wrapper)

        where N is the path index, part is one of
            feff    :      test the magnitude AND phase of F_eff (default)
            amp     :      test the total amplitude of the path
            phase   :      test the total phase of the path
            lam     :      test the mean free path
            caps    :      test the central atom phase shift
            red_fact:      test the reduction factor
            rep     :      test the real part of the complex wavenumber

        and use_wrapper is True if the test is to use the python interface to
        the feffpath library or False if the test is to use the feffrunner and
        the Feff executables.  (Currently, only feffrunner is working...)

        Sets self.rfactor (and self.rfactor_2).

        """
        if self._larch is None:
            raise Warning("cannot do path comparison -- larch broken?")

        if not self.feffran:
            print_error("You have not yet run the test Feff calculation")
            return

        if not self.available(nnnn):
            print_error("Path %d was not saved from the test Feff calculation" % nnnn)
            return

        how     = 'wrapper' if use_wrapper else 'executables'
        nnnndat = "feff%4.4d.dat" % nnnn

        blpath = feffpath(join(self.baseline, nnnndat), _larch=self._larch)
        if use_wrapper: # make the feffNNNN.dat file on the fly
            self.sp.phase_file = join(self.testrun, 'phase.pad')
            self.sp.nnnn=True
            self.sp.index=nnnn
            self.sp.verbose=self.verbose
            self.__snarf_geometry(nnnn)
            owd = getcwd()
            try:
                chdir(self.testrun)
                self.sp.calculate_xafs()
            finally:
                chdir(owd)
            n3nndat = "feff%4.4d.dat" % nnnn
            trpath = feffpath(join(self.testrun,  n3nndat), _larch=self._larch)
        else:                   # the feffNNNN.dat file was made by the monolithic feff run
            trpath = feffpath(join(self.testrun,  nnnndat), _larch=self._larch)

        if part=='feff':
            baseline_1 = getattr(blpath._feffdat, 'mag_feff') # + np.random.uniform(0,1,size=1)
            testrun_1  = getattr(trpath._feffdat, 'mag_feff')
            baseline_2 = np.sin(getattr(blpath._feffdat, 'pha_feff')) # + np.random.uniform(0,1,size=1)
            testrun_2  = np.sin(getattr(trpath._feffdat, 'pha_feff'))
            ylabel     = 'magnitude and phase'
            label      = 'magnitude'
        elif part=='amp':
            baseline_1 = getattr(blpath._feffdat, 'amp')
            testrun_1  = getattr(trpath._feffdat, 'amp')
            ylabel     = 'total amplitude'
            label      = 'amplitude'
        elif part=='phase':
            baseline_1 = np.sin(getattr(blpath._feffdat, 'pha'))
            testrun_1  = np.sin(getattr(trpath._feffdat, 'pha'))
            ylabel     = 'total phase shift'
            label      = 'phase'
        elif part=='lam':
            baseline_1 = getattr(blpath._feffdat, 'lam')
            testrun_1  = getattr(trpath._feffdat, 'lam')
            ylabel     = 'mean free path'
            label      = 'lambda'
        elif part=='phc':
            baseline_1 = getattr(blpath._feffdat, 'real_phc')
            testrun_1  = getattr(trpath._feffdat, 'real_phc')
            ylabel     = 'central atom phase shift'
            label      = 'phc'
        elif part=='red_fact':
            baseline_1 = getattr(blpath._feffdat, 'red_fact')
            testrun_1  = getattr(trpath._feffdat, 'red_fact')
            ylabel     = 'reduction factor'
            label      = 'reduction factor'
        elif part=='rep':
            baseline_1 = getattr(blpath._feffdat, 'rep')
            testrun_1  = getattr(trpath._feffdat, 'rep')
            ylabel     = 'real part of p(k)'
            label      = 'Re[p(k)]'
        else:
            if self.verbose:
                print_error("Unknown choice of parts \"%s\"\nmust be one of (feff|amp|phase|lam|phc|red_fact|rep)\nusing feff" % part)
            part       = 'feff'
            baseline_1 = getattr(blpath._feffdat, 'mag_feff')
            testrun_1  = getattr(trpath._feffdat, 'mag_feff')
            baseline_2 = np.sin(getattr(blpath._feffdat, 'pha_feff'))
            testrun_2  = np.sin(getattr(trpath._feffdat, 'pha_feff'))
            ylabel     = 'magnitude and phase'
            label      = 'magnitude'


        self.rfactor_2 = 0
        self.rfactor = sum((baseline_1 - testrun_1)**2) / sum(baseline_1**2)
        if self.verbose:
            print_warn("\nComparing %s of %s (%s) (using %s)" % (label, nnnndat, "with SCF" if self.doscf else "without SCF", how))
            self.print_geometry(blpath)
            print_warn("%s:  %f v %f " % (blpath, sum(baseline_1**2) ,  sum(testrun_1**2)))

        if self.verbose:
            print label + " R-factor = " + test_text("%.9g" % self.rfactor, (self.rfactor < self.epsilon))
        if part=='feff':
            self.rfactor_2 = sum((baseline_2 - testrun_2)**2) / sum(baseline_2**2)
            if self.verbose:
                print(" -- ",self.epsilon, sum((baseline_2 - testrun_2)**2),  sum(baseline_2**2))
                print "phase R-factor = " + test_text("%.9g" % self.rfactor_2, self.rfactor_2 < self.epsilon)

        if self.verbose: print ""

        if self.doplot:
            _newplot(blpath._feffdat.k, baseline_1, _larch=self._larch, label=label+' of baseline',
                     xlabel='wavenumber $\AA^{-1}$', ylabel=ylabel, title=nnnndat, show_legend=True, legend_loc='best')
            _plot   (trpath._feffdat.k, testrun_1,  _larch=self._larch, label=label+' of test run')
            if part=='feff':
                _plot(blpath._feffdat.k, np.gradient(baseline_2), _larch=self._larch, label='grad(phase of baseline)')
                _plot(trpath._feffdat.k, np.gradient(testrun_2),  _larch=self._larch, label='grad(phase of test run)')

        if use_wrapper and WRAPPER_AVAILABLE:
            self.sp.reset()

        if part=='feff':
            return self.rfactor < self.epsilon and self.rfactor_2 < self.epsilon
        else:
            return self.rfactor < self.epsilon


    def print_geometry(self, path):
        """
        Print out a table showing the scattering geometry of a path

        """
        print "path (reff=%.4f nlegs=%d) geometry:" % (path.reff, path.nleg)
        for atom in path.geom:
            print "\t%-2s  %7.4f  %7.4f  %7.4f  %d" % (atom[0], atom[3], atom[4], atom[5], atom[2])
        print "\t%-2s  %7.4f  %7.4f  %7.4f  %d\n" % (path.geom[0][0], path.geom[0][3], path.geom[0][4],
                                                     path.geom[0][5], path.geom[0][2])



    def s02(self, folder='testrun'):
        """
        Return the value of S02 calculated by the testrun or the baseline.
           testrun:    larch> group.s02()
           baseline:   larch> group.s02('baseline')
        """
        if folder == 'testrun':
            chidat = join(self.testrun, 'chi.dat')
        else:
            chidat = join(self.baseline, 'chi.dat')

        with open(chidat, 'r') as searchfile:
            for line in searchfile:
                m = re.search('S02=(\d\.\d+)', line)
                if m:
                    return float(m.group(1))
        return -1


    def radii(self, folder='testrun', radius='muffintin'):
        """
        Return a list of the selected radii for each unique potential
           larch> group.s02(folder, radius)
        where folder = testrun|baseline and radius = norman|muffintin

        """
        if folder == 'testrun':
            chidat = join(self.testrun, 'chi.dat')
        else:
            chidat = join(self.baseline, 'chi.dat')

        values = list()
        with open(chidat, 'r') as searchfile:
            for line in searchfile:
                if radius == 'norman':
                    m = re.search('Rnm=\s*(\d\.\d+)', line)
                else:
                    m = re.search('Rmt=\s*(\d\.\d+)', line)
                if m:
                    values.append(float(m.group(1)))
        return values


    def feffterms(self, nnnn=1):
        """
        Fetch the various numbers in the feffNNNN.dat header for both
        the baseline and testrun.  A table of these values is printed
        if the verbose flag is True.  Return True if all are equal.

        """
        if (not self.feffran):
            #if self.verbose: print colored("You have not yet made the test run of Feff", 'magenta', attrs=['bold'])
            raise Exception("You have not yet made the test run of Feff")
        nnnndat = "feff%4.4d.dat" % nnnn
        bl      = feffpath(join(self.baseline, nnnndat), _larch=self._larch)
        tr      = feffpath(join(self.testrun,  nnnndat), _larch=self._larch)

        termdict = {'edge':   'energy threshold relative to atomic value',
                    'gam_ch': 'core level energy width',
                    'kf':     'k value at Fermi level',
                    'mu':     'Fermi level in eV',
                    'rs_int': 'interstitial radius',
                    'vint':   'interstitial potential'}
        if self.verbose:
            print "%-8s %-42s   %10s  %10s  %s" % ('key', 'description', 'baseline', 'testrun', 'same?')
            print "-----------------------------------------------------------------------------------"
        ok = True
        for key in termdict:
            same = abs(getattr(bl._feffdat, key) - getattr(tr._feffdat, key))/getattr(bl._feffdat, key) < self.epsilon
            if self.verbose: print "%-6s   %-42s : %10s  %10s  %s" % (key, termdict[key], getattr(bl._feffdat, key),
                                                                      getattr(tr._feffdat, key), same)
            ok = ok and same
        return ok



    def fit(self):
        """
        Perform a canned fit using the baseline and testrun Feff
        calculations.

        Sets self.blfit and self.trfit, containing the feffit groups
        for the fits suing the baseline and test run Feff
        calculations, respectively.

        """
        sys.path.append(self.path)
        module = importlib.import_module(self.folder, package=None)

        print "\tfitting %s using %s" % (self.folder, 'baseline')
        self.blfit = module.do_fit(self, 'baseline')
        print "\tfitting %s using %s" % (self.folder, 'testrun')
        self.trfit = module.do_fit(self, 'testrun')


    def fitcompare(self):
        """
        Perform a canned fit using a sequence of Feff calculations

        This is not used for unit testing feff85exafs.  It is a
        generalization of the fit() method intended for use with a
        generic testing framework, e.g. https://github.com/bruceravel/SCFtests

        Uses self.fittest, the name of a folder containing a sequence of
        Feff calculations.

        Sets self.models, a list of subfolders containing the identifying
        strings of the sequence of calculations.

        Returns a bunch of self.<id>, where <id> are feffit groups
        containing the fits in the sequence.
        """
        sys.path.append(self.folder)
        module = importlib.import_module(self.folder, package=None)
        save = self.doscf

        if self.fittest is None:
            raise Exception("You must select a test folder.")

        self.models = []
        for d in sorted(listdir(join(self.folder, self.fittest))):
            if isdir(join(self.folder, self.fittest, d)):
                print '>>>>>>>>> fitting with model: %s' % d
                self.models.append(d)
                this = module.do_fit(self, d, firstshell=self.firstshell, fittest=self.fittest)
                setattr(self, d, this)


    def clean(self):
        """
        Remove testrun folder
        """
        if isdir(self.testrun): rmtree(self.testrun)
        self.feffran = False


######################################################################

def ut(folder=None, _larch=None, **kws):
    """
    Make a Feff85exafsUnitTestGroup group given a folder containing a
    baseline calculation

    """
    return Feff85exafsUnitTestGroup(folder=folder, _larch=_larch)

def ir(folder=None, _larch=None, **kws):
    """
    _i_mport and _r_un

    """
    utobj=ut(folder)
    utobj.run()
    return utobj

def irc(folder=None, _larch=None, **kws):
    """
    _i_mport, _r_un, and _c_ompare

    """
    utobj=ut(folder)
    utobj.run()
    utobj.compare(1)
    return utobj

def irf(folder=None, _larch=None, **kws):
    """_i_mport, _r_un, and _f_it

    """
    utobj=ut(folder)
    utobj.run()
    utobj.fit()
    return utobj

def registerLarchPlugin(): # must have a function with this name!
    return ('f85ut', { 'ut': ut, 'ir': ir, 'irc': irc, 'irf': irf })



#!/usr/bin/env python
"""A larchified, simplified, feature-full wrapper around the thin
(but weird) SWIG wrapper for the feffpath library.

Use of this plugin presumes that the feffpath C library is installed
and that the SWIG wrapper has been built, compiled, and installed.

  add_plugin("scatteringpath")
  a = scatteringpath()
  try:
     a.index = 1
  except ValueError:
     print "Bad index value, index still %" : a.index
  #endtry
  try:
     a.degen = 48
  except ValueError:
     print "Bad degeneracy value, degen still %" : a.degen
  #endtry
  a.atom(1.805,0,1.805,1)
  if a.errorcode:
     print a.errormessage
     exit()
  #endif
  a.make()
  if a.errorcode:
     print a.errormessage
     exit()
  #endif
  a.clear()

Methods:

  atom:   add a scattering atom to a path and increment nleg
  make:   compute F_eff for a scattering path
  clear:  reinitialize  the scattering path

Attributes (input):

  phpad        : string   path to `phase.pad`                       `phase.pad`       
  index        : integer  path index                                9999              
  degen        : float    path degeneracy                           required input    
  nleg         : integer  number of legs in path                    set using atom method (read-only)
  rat          : list of tuples  cartesian positions of atoms in path      set using atom method
  ipot         : list of int     unique potentials of atoms in path        set using atom method
  iorder       : integer  order of approximation in genfmt          2
         
  nnnn         : boolean  flag to write `feffNNNN.dat` file         False             
  json         : boolean  flag to write `feffNNNN.json` file        False             
  verbose      : boolean  flag to write screen messages             False             

  ipol         : boolean  flag to do polarization calculation       False (set True when evec|xivec|elpty set)
  evec         : array    polarization vector                       (0,0,0)           
  elpty        : float    ellipticity                               0                 
  xivec        : array    direction of X-ray propagation            (0,0,0)           

Attributes (output):

  ri           : array    leg lengths                                                
  beta         : array    beta angles                                                
  eta          : array    eta angles                                                 
  reff         : float    half path length (computed from ri)

  ne           : integer  number of energy points actually used by Feff (read-only)
  k            : array    k grid for feff path calculation, column 1 in `feffNNNN.dat`
  real_phc     : array    central atom phase shifts. column 2 in `feffNNNN.dat`
  mag_feff     : array    magnitude of F_eff, column 3 in `feffNNNN.dat`
  pha_feff     : array    phase of F_eff, column 4 in `feffNNNN.dat`
  red_fact     : array    reduction factor, column 5 in `feffNNNN.dat`
  lam          : array    mean free path, column 6 in `feffNNNN.dat`
  rep          : array    real part of complex momentum, column 7 in `feffNNNN.dat`

Output attributes related to Larch's feffdat Group
  geom         : list     path geometry: list of (Symbol, Z, ipot, x, y, z)
  amp          : array    mag_feff * red_fact
  pha          : array    pha_feff + real_phc

Error handling attributes
  errorcode    : integer  error code from `atom` or `make`
  errormessage : string   error message from `atom` or `make`

The following attributes are related to Feff's potential model
  edge         : float    energy threshold relative to atomic value (a poor estimate)
  gam_ch       : float    core level energy width
  kf           : float    k value at Fermi level
  mu           : float    Fermi level, eV
  rnorman      : float    Norman radius
  version      : string   Feff version
  exch         : string   brief description of the electronic exchange model
  rs_int       : float    interstitial radius
  vint         : float    interstitial potential
  version      : string   the version of feff and the feffpath revision

This attribute of Larch's _feffdat group is not captured
  potentials   : list     path potentials: list of (ipot, z, r_MuffinTin, r_Norman)

"""

# LICENSE AND COPYRIGHT
#
# To the extent possible, the authors have waived all rights granted by
# copyright law and related laws for the code and documentation that
# make up the Python Interface to the feffpath library.  While information
# about Authorship may be retained in some files for historical reasons,
# this work is hereby placed in the Public Domain.  This work is
# published from: United States.
#
# Note that the onepath and feffpath libraries themselves are NOT
# public domain, nor is the Fortran source code for Feff that it
# relies upon.
#
# Author: Bruce Ravel (bravel AT bnl DOT gov).
# Last update: 12 December, 2014

import larch
from larch import (Group, Parameter, isParameter, ValidateLarchPlugin, param_value,
                   use_plugin_path, isNamedClass, Interpreter)
use_plugin_path('xafs')
use_plugin_path('xray')
from   xraydb_plugin import atomic_symbol

import sys
from   os            import name
from   os.path       import isfile
from   numpy         import array

#installdir = larch.site_config.larchdir
installdir = larch.larchlib.larchdir
dlldir     = larch.larchlib.get_dlldir()


dllfile=installdir+'/dlls/'+dlldir
if not dllfile in sys.path:
    sys.path.append(dllfile)
moddir=installdir+'/modules'
if not moddir in sys.path:
    sys.path.append(moddir)

import feffpathwrapper



class FeffPathBoolean(object):
    """A descriptor for boolean-valued FeffPath attributes"""
    def __init__(self, name):
        self._name = name

    def __get__(self, instance, owner):
        return getattr(instance.wrapper, self._name)

    def __set__(self, instance, val):
        setattr(instance.wrapper, self._name, val)

class FeffPathUnsettableInteger(object):
    """A descriptor for integer-valued FeffPath attributes which are not
    intended to be set by the user, instead are set by calls to the
    atom or make methods.

    """
    def __init__(self, name):
        self._name = name

    def __get__(self, instance, owner):
        return int(getattr(instance.wrapper, self._name))

    def __set__(self, instance, val):
        pass

class FeffPathUnsettableString(object):
    """A descriptor for string-valued FeffPath attributes which are not
    intended to be set by the user, instead are set by calls to the
    make method.

    """
    def __init__(self, name):
        self._name = name

    def __get__(self, instance, owner):
        return str(getattr(instance.wrapper, self._name))

    def __set__(self, instance, val):
        pass

class FeffPathUnsettableFloat(object):
    """A descriptor for float-valued FeffPath attributes which are not
    intended to be set by the user, instead are set by calls to the
    make method.

    """
    def __init__(self, name):
        self._name = name

    def __get__(self, instance, owner):
        return float(getattr(instance.wrapper, self._name))

    def __set__(self, instance, val):
        pass

class FeffPathNlegList(object):
    """A descriptor for list-valued FeffPath attributes which have nleg
    members and describe the geometry of the path, i.e. ri, beta, and
    eta

    """
    def __init__(self, name):
        self._name = name

    def __get__(self, instance, owner):
        lst=[]
        for i in range(instance.wrapper.nleg):
            x = eval('feffpathwrapper.get_'+self._name+'(instance.wrapper,i)')
            lst.append(float(x))
            #lst.append(feffpathwrapper.get_ri(self.wrapper,i))
        return lst

    def __set__(self, instance, val):
        pass

class FeffPathColumn(object):
    """A descriptor for the array-valued FeffPath attributes which are the
    columns of feffNNNN.dat"""
    def __init__(self, name):
        self._name = name

    def __get__(self, instance, owner):
        arr=[]
        for i in range(instance.wrapper.ne):
            x = eval('feffpathwrapper.get_'+self._name+'(instance.wrapper,i)')
            arr.append(float(x))
            #arr.append(feffpathwrapper.get_k(instance.wrapper,i))
        return array(arr)

    def __set__(self, instance, val):
        pass



class FeffPath(Group):
    """
    A larchified, simplified, feature-full wrapper around the thin
    (but weird) SWIG wrapper for the feffpath library.
    """

    nnnn      = FeffPathBoolean('nnnn')
    json      = FeffPathBoolean('json')
    verbose   = FeffPathBoolean('verbose')
    ipol      = FeffPathBoolean('ipol')

    nleg      = FeffPathUnsettableInteger('nleg')
    ne        = FeffPathUnsettableInteger('ne')
    errorcode = FeffPathUnsettableInteger('errorcode')

    exch      = FeffPathUnsettableString('exch')
    version   = FeffPathUnsettableString('version')
    errormessage = FeffPathUnsettableString('errormessage')

    edge      = FeffPathUnsettableFloat('edge')
    gam_ch    = FeffPathUnsettableFloat('gam_ch')
    kf        = FeffPathUnsettableFloat('kf')
    mu        = FeffPathUnsettableFloat('mu')
    rnorman   = FeffPathUnsettableFloat('rnorman')
    rs_int    = FeffPathUnsettableFloat('rs_int')
    vint      = FeffPathUnsettableFloat('vint')

    ri        = FeffPathNlegList('ri')
    beta      = FeffPathNlegList('beta')
    eta       = FeffPathNlegList('eta')

    k         = FeffPathColumn('k')
    real_phc  = FeffPathColumn('real_phc')
    mag_feff  = FeffPathColumn('mag_feff')
    pha_feff  = FeffPathColumn('pha_feff')
    red_fact  = FeffPathColumn('red_fact')
    lam       = FeffPathColumn('lam')
    rep       = FeffPathColumn('rep')

    def __init__(self, folder=None, _larch=None, **kws):
        kwargs = dict(name='FeffPath wrapper')
        kwargs.update(kws)
        Group.__init__(self,  **kwargs)
        self._larch     = Interpreter()
        self.wrapper    = feffpathwrapper.FEFFPATH()
        feffpathwrapper.create_path(self.wrapper)
        self.wrapper.phpad = ''
        self.ipot = []
        self.rat  = []
        self.geom = []


    ## ---- scalar valued attributs, these get their own properties so
    ##      tailored ValueErrors can be issued

    @property
    def phpad(self):
        return self.wrapper.phpad
    @phpad.setter
    def phpad(self,value):
        if not isfile(value):
            raise ValueError("%s is not a readable file" % value)
        self.wrapper.phpad = value

    @property
    def index(self):
        return int(self.wrapper.index)
    @index.setter
    def index(self,value):
        value = int(value)
        if value < 1 or value > 9999:
            raise ValueError("Index values must be between 1 and 9999: %s" % value)
        self.wrapper.index = long(value)

    @property
    def degen(self):
        return self.wrapper.degen
    @degen.setter
    def degen(self,value):
        if value < 1:
            raise ValueError("Negative degeneracies not allowed: %s" % value)
        self.wrapper.degen = value

    @property
    def iorder(self):
        return int(self.wrapper.iorder)
    @iorder.setter
    def iorder(self,value):
        if value < 0 or value > 10:
            raise ValueError("Iorder value must be 10 or less: %s" % value)
        self.wrapper.iorder = long(value)

    @property
    def elpty(self):
        return self.wrapper.elpty
    @elpty.setter
    def elpty(self,value):
        if value < 0:
            raise ValueError("elpty must be between 0 and 1: %s" % value)
        elif value > 1:
            raise ValueError("elpty must be between 0 and 1: %s" % value)
        self.wrapper.elpty = value
        self.wrapper.ipol  = True

    ## ---- error handling (this is the only string-valued attribute, it is returned from atom and make)

    # @property
    # def errormessage(self):
    #     return self.wrapper.errormessage
    # @errormessage.setter
    # def errormessage(self,value):
    #     pass


    ## ---- 3-vec valued attributes (with tailored ValueErrors)

    @property
    def evec(self):
        return [feffpathwrapper.get_evec(self.wrapper,0),
                feffpathwrapper.get_evec(self.wrapper,1),
                feffpathwrapper.get_evec(self.wrapper,2)]
    @evec.setter
    def evec(self, vec):
        if type(vec).__name__ not in ('list', 'tuple'):
            raise ValueError("evec must be 3 element list (or tuple) with (x,y,z) of the electric vector (not a list/tuple)")
        elif len(vec) != 3:
            raise ValueError("evec must be 3 element list (or tuple) with (x,y,z) of the electric vector (not of length 3)")
        feffpathwrapper.set_evec(self.wrapper,0,vec[0])
        feffpathwrapper.set_evec(self.wrapper,1,vec[1])
        feffpathwrapper.set_evec(self.wrapper,2,vec[2])
        self.wrapper.ipol = True

    @property
    def xivec(self):
        return [feffpathwrapper.get_xivec(self.wrapper,0),
                feffpathwrapper.get_xivec(self.wrapper,1),
                feffpathwrapper.get_xivec(self.wrapper,2)]
    @xivec.setter
    def xivec(self, vec):
        if type(vec).__name__ not in ('list', 'tuple'):
            raise ValueError("xivec must be 3 element list/tuple with (x,y,z) of the Poynting vector (not a list/tuple)")
        elif len(vec) != 3:
            raise ValueError("xivec must be 3 element list/tuple with (x,y,z) of the Poynting vector (not of length 3)")
        feffpathwrapper.set_xivec(self.wrapper,0,vec[0])
        feffpathwrapper.set_xivec(self.wrapper,1,vec[1])
        feffpathwrapper.set_xivec(self.wrapper,2,vec[2])
        self.wrapper.ipol = True


    @property
    def iz(self):
        lst = [];
        for i in range(feffpathwrapper.nphx):
            lst.append(int(feffpathwrapper.get_iz(self.wrapper,i)))
        return lst
    @iz.setter
    def iz(self, vec):
        pass

    @property
    def amp(self):
        return self.mag_feff * self.red_fact
    @amp.setter
    def amp(self, value):
        pass

    @property
    def pha(self):
        return self.pha_feff + self.real_phc
    @pha.setter
    def pha(self, value):
        pass


    ## ---- methods

    def atom(self, x, y, z, ip):
        """
        Add a scatterer to a scattering path and increment nleg.

          a = scatteringpath()
          a.atom(0, 0, -3.61, 1)

        Arguments are x,y,z coordinates of the scatterer and it ipot value

        The Cartesian coordinates are relative to the ABSORBER AT 0,0,0
        """
        self.ipot.append(ip)
        self.rat.append((x, y, z))
        feffpathwrapper.add_scatterer(self.wrapper, x, y, z, ip)

    def make(self):
        """
        Compute F_eff for a path, make columns of feffNNNN.dat available as
        properties of the object.  Write feffNNNN.dat file if verbose attribute
        set to True.

          a = scatteringpath()
          a.atom(0, 0, -3.61, 1)
          a.make()

        """
        feffpathwrapper.make_path(self.wrapper)
        if self.errorcode == 0:
            self.geom = []
            for i in range(self.nleg-1):
                ip = self.ipot[i]
                this = (str(atomic_symbol(self.iz[ip], _larch=self._larch)),
                        self.iz[ip], ip, self.rat[i][0], self.rat[i][1], self.rat[i][2])
                self.geom.append(this)

    def clear(self):
        """
        Reinitialize a path.

          a.clear()

        """
        feffpathwrapper.clear_path(self.wrapper)
        self.ipot = []
        self.rat  = []
        self.geom = []



######################################################################

def scatpath(_larch=None, **kws):
    """
    Make and return an empty FeffPath group
    """
    return FeffPath(_larch=_larch)

def registerLarchPlugin(): # must have a function with this name!
    return ('_xafs', { 'scatteringpath': scatpath })



## feff85exafs build system based on scons
## see HEADERS/license.h for license information

import SCons

from SCons.Environment import Environment
import sys
import os
if os.name == 'nt':
    from os import getcwd
else:
    from os import getcwd, chown
from os.path import realpath, join

from SCons.Script.SConscript import SConsEnvironment
if os.name != 'nt':
    from pwd import getpwnam, getpwuid
    from grp import getgrnam, getgrgid

DEBUG_ENV = False

## jsondir:
##   -I or -module flags: see
##   http://stackoverflow.com/questions/8855896/specify-directory-where-gfortran-should-look-for-modules
##   n.b.: this gets evaluated in one of the subfolders, hence the ..

def CompilationEnvironment():
    """
    Determine how to build the Fortran parts of feff85exafs
    """
    args = {}
    jsondir = realpath(join(getcwd(), '..', 'json-fortran'))
    flags = {'jsondir': jsondir}

    gfortran_comp_flags = '-O3 -ffree-line-length-none -g -Wall'
    gfortran_link_flags = None

    json_comp_flags = ' -I{jsondir} -J{jsondir}'.format(**flags)
    json_link_flags = ' -L{jsondir} -ljsonfortran'.format(**flags)

    # darwin:
    if os.name == 'posix' and sys.platform=='darwin':
        args['platform'] = 'darwin'
        gfortran_comp_flags = '-O2 -arch x86_64 -Wall'
        gfortran_link_flags = '-dynamiclib -L/usr/local/gfortran/lib/ -lgfortran -lgfortranbegin'

    # windows: needs work!
    if os.name  == 'nt':
        #args['platform'] = 'Windows'
        pass

    env = Environment(**args)

    if env['FORTRAN'] == 'gfortran':
        # this was the suggestion in the top level Makefile in what
        # the FP gave us: "-O3 -ffree-line-length-none -finit-local-zero"
        # -O3 makes sense, as does -finit-local-zero.  I don't understand
        # the advantage of -ffree-line-length-none -- it seems to
        # encourage poor code style -- like we need more of that!
        # also,  -finit-local-zero fails on FMS/fmstot.f
        env = Environment(FORTRANFLAGS = gfortran_comp_flags + json_comp_flags, CFLAGS = '-g')
        if gfortran_link_flags is not None:
            env.Replace(SHLINKFLAGS = gfortran_link_flags + json_link_flags)
    elif env['FORTRAN'] == 'g77':
        env = Environment(FORTRANFLAGS = '-Wall -O2')
    elif env['FORTRAN'] == 'xlf':
        env = Environment(FORTRANFLAGS = '-qextern=trap')
    elif env['FORTRAN'] == 'ifort':
        ## I think the -module flg is correct ... untested ...
        env = Environment(FORTRANFLAGS = '-O3 -module '+jsondir)

    if DEBUG_ENV:
        for key, val in env.items():
            try:
                print( key, val)
            except:
                pass
        sys.exit()

    if os.name == 'nt':
        env.PrependENVPath('PATH', os.environ['PATH'])
    
    return env

## need to be able to get prefix from command line
def InstallEnvironment():
    """
    Determine installation locations for libraries and executables.
    """
    ienv = Environment()
    #prefix = ARGUMENTS.get('prefix', '/usr/local')
    if os.name == 'nt':
        import larch
        prefix = larch.larchlib.sys_larchdir
        dlldir = larch.larchlib.get_dlldir()
        # Here are our installation paths:
        ienv['i_prefix'] = prefix
        ienv['i_lib']    = join(prefix, 'dlls', dlldir)
        ienv['i_bin']    = join(prefix, 'bin')
        ienv['i_inc']    = join(prefix, 'include')
        ienv['i_data']   = join(prefix, 'share')
    else:
        prefix = '/usr/local'
        # Here are our installation paths:
        ienv['i_prefix'] = prefix
        ienv['i_lib']    = join(prefix, 'lib')
        ienv['i_bin']    = join(prefix, 'bin')
        ienv['i_inc']    = join(prefix, 'include')
        ienv['i_data']   = join(prefix, 'share')
    return ienv


def FindOtherObjects(deplist, env, kind):
    objsuff = env['SHOBJSUFFIX']
    if kind == 'static': objsuff = env['OBJSUFFIX']
    out = []
    for dep in deplist:
        dname, fname = dep.split('/')
        out.append(join('..', dname, fname + objsuff))
    return out

##+----------------------------------------------------------------------------------------------------
## This implementation of a Chown factory closely follows Chmod from /usr/lib/scons/SCons/Defaults.py
## and the discussion from http://www.scons.org/wiki/InstallTargets

def get_paths_str(dest):
    # If dest is a list, we need to manually call str() on each element
    if SCons.Util.is_List(dest):
        elem_strs = []
        for element in dest:
            elem_strs.append('"' + str(element) + '"')
        return '[' + ', '.join(elem_strs) + ']'
    else:
        return '"' + str(dest) + '"'

## FIXME.MAYBE: recognize numeric and string uid and gid
def chown_func(dest, owner, group):
    SCons.Node.FS.invalidate_node_memos(dest)
    if not SCons.Util.is_List(dest):
        dest = [dest]
    for element in dest:
        #os.chmod(str(element), getpwnam(owner).pw_uid, getgrnam(owner).gr_gid)
        chown(str(element), owner, owner)

def chown_strfunc(dest, owner, group):
    return 'Chown(%s, %s.%s)' % (get_paths_str(dest), getpwuid(owner).pw_name, getgrgid(group).gr_name)

SConsEnvironment.Chown = SCons.Action.ActionFactory(chown_func, chown_strfunc)

def InstallOwner(env, dest, files, owner, group):
    """
    Used for files to be owned by the normal user even if the installation is run as sudo
    """
    obj = env.Install(dest, files)
    for i in obj:
        env.AddPostAction(i, env.Chown(str(i), owner, group))
    return dest

SConsEnvironment.InstallOwner = InstallOwner



import numpy as np
from feffpath import ScatteringPath

path = ScatteringPath(phase_file='phase.pad')
assert(path is not None)

path.set_absorber( x=0.01,   y=0.1,   z=0.01)
path.add_scatterer(x=1.8058, y=0.005, z=1.8063, ipot=1)
path.degen = 12


path.calculate_xafs()

assert(path.nleg == 2)
assert(path.iz[0] == 29)
assert(path.rat[0, 0] > 0.00)
assert(path.rat[0, 0] < 0.20)

assert(path.rat[0, 1] > 1.70)
assert(path.rat[0, 1] < 1.90)

assert(path.rnorman > 2.50)
assert(path.rnorman < 2.80)

assert(path.rs   > 1.80)
assert(path.rs   < 2.10)
assert(path.kf   > 0.80)
assert(path.kf   < 1.10)

assert(path.exch_label.startswith('H-L'))


npts = 1 + max(np.where(path.kfeff > 0)[0])
assert(npts > 25)

assert(path.kfeff[5] > 0.490)
assert(path.kfeff[5] < 0.510)

assert(path.rep[5] > 1.880)
assert(path.rep[5] < 1.900)

assert(path.lam[5] > 16.4)
assert(path.lam[5] < 16.6)

assert(path.real_phc[5] > 3.92)
assert(path.real_phc[5] < 3.98)






from os.path import realpath, exists, join

from larch import (Group, Parameter, isParameter, param_value, use_plugin_path, isNamedClass)
use_plugin_path('io')
from xdi import read_xdi
use_plugin_path('xafs')
from feffit import feffit_dataset, feffit_transform, feffit, feffit_report
from feffdat import feffpath
use_plugin_path('wx')
from plotter import (_newplot, _plot)


def do_fit(self, which):

    if which == 'testrun':
        folder = self.testrun
    else:
        folder = self.baseline

    data = read_xdi(join(self.path, 'UO2.chik'), _larch=self._larch)

    gds = Group(amp    = Parameter(1,      vary=True, _larch=self._larch),
                enot   = Parameter(0.01,   vary=True, _larch=self._larch),
                sso    = Parameter(0.003,  vary=True, _larch=self._larch),
                ssu    = Parameter(0.003,  vary=True, _larch=self._larch),
                sso2   = Parameter(0.003,  vary=True, _larch=self._larch),
                dro    = Parameter(0.0001, vary=True, _larch=self._larch),
                dru    = Parameter(0.0001, vary=True, _larch=self._larch),
                dro2   = Parameter(0.0001, vary=True, _larch=self._larch),
                nu     = Parameter(12,     vary=True, _larch=self._larch),
                no2    = Parameter(expr='2*nu',       _larch=self._larch),
                _larch=self._larch  )

    paths = list() 
    paths.append(feffpath(realpath(join(folder, "feff0001.dat")), # 1st shell O SS
                          s02    = 'amp',
                          e0     = 'enot',
                          sigma2 = 'sso',
                          deltar = 'dro', _larch=self._larch))
    # paths.append(feffpath(realpath(join(folder, "feff0002.dat")), # triangle in first shell
    #                       s02    = 'amp',
    #                       e0     = 'enot',
    #                       sigma2 = 'sso*1.5',
    #                       deltar = 'dro*(1+sqrt(2))/2', _larch=self._larch))
    paths.append(feffpath(realpath(join(folder, "feff0003.dat")), # 2nd shell U SS
                          degen  = 1,
                          s02    = 'amp*nu',
                          e0     = 'enot',
                          sigma2 = 'ssu',
                          deltar = 'dru', _larch=self._larch))
    # paths.append(feffpath(realpath(join(folder, "feff0004.dat")), # 1st shell, longer triangle
    #                       s02    = 'amp',
    #                       e0     = 'enot',
    #                       sigma2 = '2*sso',
    #                       deltar = '2*dro', _larch=self._larch))
    paths.append(feffpath(realpath(join(folder, "feff0006.dat")), # 3rd shell O SS
                          degen  = 1,
                          s02    = 'amp*no2',
                          #s02    = 'amp*nu*2',
                          e0     = 'enot',
                          sigma2 = 'sso2',
                          deltar = 'dro2', _larch=self._larch))
    paths.append(feffpath(realpath(join(folder, "feff0007.dat")), # 1st shell, non-forward linear through absorber
                          s02    = 'amp',
                          e0     = 'enot',
                          sigma2 = 'sso2',
                          deltar = 'dro2', _larch=self._larch))
    paths.append(feffpath(realpath(join(folder, "feff0008.dat")), # 1st shell forward through absorber
                          s02    = 'amp',
                          e0     = 'enot',
                          sigma2 = '2*sso',
                          deltar = '2*dro', _larch=self._larch))
    paths.append(feffpath(realpath(join(folder, "feff0009.dat")), # rattle in 1st shell
                          s02    = 'amp',
                          e0     = 'enot',
                          sigma2 = '2*sso',
                          deltar = '2*dro', _larch=self._larch))


    trans = feffit_transform(kmin=3, kmax=11, kw=(2,1,3), dk=1, window='hanning', rmin=1.25, rmax=4.3, _larch=self._larch)
    dset  = feffit_dataset(data=data, pathlist=paths, transform=trans, _larch=self._larch)
    fit   = feffit(gds, dset, _larch=self._larch)

    if self.doplot:
        offset = max(dset.data.chir_mag)
        _newplot(dset.data.r,  dset.data.chir_mag+offset, xmax=8, win=2,
              xlabel=r'$R \rm\,(\AA)$', label='data',
              ylabel=r'$|\chi(R)| \rm\,(\AA^{-3})$',
              title='Fit to '+self.folder, show_legend=True, _larch=self._larch)
        _plot(dset.model.r, dset.model.chir_mag+offset, label='fit', win=2, _larch=self._larch)
        _plot(dset.data.r,  dset.data.chir_re, label='data', win=2, _larch=self._larch)
        _plot(dset.model.r, dset.model.chir_re, label='fit', win=2, _larch=self._larch)
    #end if
    
    if self.verbose:
        print feffit_report(fit, _larch=self._larch)
    #end if

    return fit
#end def



import sys

from os.path import realpath, exists, join

from larch import (Group, param_group, Parameter, isParameter, param_value, use_plugin_path, isNamedClass)
use_plugin_path('io')
from xdi import read_xdi
use_plugin_path('xafs')
from feffit import feffit_dataset, feffit_transform, feffit, feffit_report
from feffdat import feffpath
use_plugin_path('wx')
from plotter import (_newplot, _plot)


def do_fit(self, which):

    if which == 'testrun':
        folder = self.testrun
    elif which == 'baseline':
        folder = self.baseline
    else:
        folder = realpath(join(self.folder, 'baseline', which))
    #endif

    print '>>>>>> %s' % folder

    data = read_xdi(join(self.path, 'Copper.chik'), _larch=self._larch)

    gds = param_group(amp    = Parameter(1,     vary=True),
                enot   = Parameter(1e-7,  vary=True),
                thetad = Parameter(500,   vary=True),
                temp   = Parameter(10,    vary=False),
                alpha  = Parameter(1e-7,  vary=True),
                ss1    = Parameter(0.003, vary=True),
                _larch=self._larch  )

    paths = list()
    for index in range(1, 14):
        nnnn = realpath(join(folder, "feff%4.4d.dat" % index))
        if not exists(nnnn):
            continue
        #end if
        if index > 1:
            sigsqr = 'sigma2_debye(temp, thetad)'
        else:
            sigsqr = 'ss1'
        #end if

        paths.append(feffpath(nnnn, s02='amp', e0='enot',
                              sigma2=sigsqr,
                              deltar='alpha*reff',
                              _larch=self._larch))

    #end for

    trans = feffit_transform(kmin=3, kmax=16, kw=(2,1,3), dk=1, window='hanning', rmin=1.7, rmax=5.1, _larch=self._larch)
    dset  = feffit_dataset(data=data, pathlist=paths, transform=trans, _larch=self._larch)

    fit   = feffit(gds, dset, _larch=self._larch)

    if self.doplot:
        offset = max(dset.data.chir_mag)
        _newplot(dset.data.r,  dset.data.chir_mag+offset, xmax=8, win=2,
              xlabel=r'$R \rm\,(\AA)$', label='data',
              ylabel=r'$|\chi(R)| \rm\,(\AA^{-3})$',
              title='Fit to '+self.folder, show_legend=True, _larch=self._larch)
        _plot(dset.model.r, dset.model.chir_mag+offset, label='fit', win=2, _larch=self._larch)
        _plot(dset.data.r,  dset.data.chir_re, label='data', win=2, _larch=self._larch)
        _plot(dset.model.r, dset.model.chir_re, label='fit', win=2, _larch=self._larch)
    #end if

    if self.verbose:
        print feffit_report(fit, _larch=self._larch)
    #end if

    return fit
#end def






from os.path import realpath, exists, join

from larch import (Group, Parameter, use_plugin_path)
use_plugin_path('io')
from xdi import read_xdi
use_plugin_path('xafs')
from feffit import feffit_dataset, feffit_transform, feffit, feffit_report
from feffdat import feffpath, _ff2chi
use_plugin_path('wx')
from plotter import (_newplot, _plot)

def do_fit(self, which):

    if which == 'testrun':
        folder = self.testrun
    else:
        folder = self.baseline
    #end if

    data = read_xdi(join(self.path, 'bromoadamantane.chik'), _larch=self._larch)

    gds = Group(amp     = Parameter(1.021,     vary=True,  _larch=self._larch),
                enot    = Parameter(4.01,      vary=True,  _larch=self._larch),
                delr    = Parameter(-0.007,    vary=True,  _larch=self._larch),
                brc     = Parameter(expr = '1.9521+delr',  _larch=self._larch),
                ss      = Parameter(0.003,     vary=True,  _larch=self._larch),
                phir    = Parameter(109.29960 * 3.141592653589793 / 180,   vary=False, _larch=self._larch),
                cc      = Parameter(1.53780,   vary=False, _larch=self._larch),
                tanbeta = Parameter(expr = '(brc+cc)*tan(phir/2) / (brc-cc)', _larch=self._larch),
                beta    = Parameter(expr = 'atan(tanbeta)', _larch=self._larch),
                brc2    = Parameter(expr = '(brc-cc)*cos(phir/2)/cos(beta)', _larch=self._larch),
                drh     = Parameter(0.04,      vary=True,  _larch=self._larch),
                ssh     = Parameter(0.005,     vary=True,  _larch=self._larch),
                ss2     = Parameter(expr = 'ss*(brc2/brc)**2', _larch=self._larch),
                c3      = Parameter(-0.0007,   vary=True,  _larch=self._larch),
                _larch=self._larch  )

    paths = list()
    paths.append(feffpath(realpath(join(folder, "feff0001.dat")),
                          s02    = 'amp',
                          e0     = 'enot',
                          sigma2 = 'ss',
                          deltar = 'delr',
                          third  = 'c3', _larch=self._larch))
    paths.append(feffpath(realpath(join(folder, "feff0002.dat")),
                          s02    = 'amp',
                          e0     = 'enot',
                          sigma2 = 'ss2',
                          deltar = 'brc2-2.8565', _larch=self._larch))
    paths.append(feffpath(realpath(join(folder, "feff0003.dat")),
                          s02    = 'amp',
                          e0     = 'enot',
                          sigma2 = 'ssh',
                          deltar = 'drh', _larch=self._larch))
    paths.append(feffpath(realpath(join(folder, "feff0004.dat")),
                          s02    = 'amp',
                          e0     = 'enot',
                          sigma2 = '(ss+ss2)/2',
                          deltar = '(brc+brc2+cc)/2 - 3.173', _larch=self._larch))
    paths.append(feffpath(realpath(join(folder, "feff0005.dat")),
                          s02    = 'amp',
                          e0     = 'enot',
                          sigma2 = '(ss+ss2)/2',
                          deltar = '(brc+brc2+cc)/2 - 3.173', _larch=self._larch))


    trans = feffit_transform(kmin=2.5, kmax=13, kw=(2,1,3), dk=1, window='hanning', rmin=1.25, rmax=3, _larch=self._larch)
    dset  = feffit_dataset(data=data, pathlist=paths, transform=trans, _larch=self._larch)
    fit   = feffit(gds, dset, _larch=self._larch)

    if self.doplot:
        offset = max(dset.data.chir_mag)
        _newplot(dset.data.r,  dset.data.chir_mag+offset, xmax=8, win=2,
                 xlabel=r'$R \rm\,(\AA)$', label='data',
                 ylabel=r'$|\chi(R)| \rm\,(\AA^{-3})$',
                 title='Fit to '+self.folder, show_legend=True, _larch=self._larch)
        _plot(dset.model.r, dset.model.chir_mag+offset, label='fit', win=2, _larch=self._larch)
        _plot(dset.data.r,  dset.data.chir_re, label='data', win=2, _larch=self._larch)
        _plot(dset.model.r, dset.model.chir_re, label='fit', win=2, _larch=self._larch)
    #end if
    
    if self.verbose:
        print feffit_report(fit, _larch=self._larch)
    #end if

    return fit
#end def



#!/usr/bin/python
"""
Feff85L EXAFS Scatternig Path for Python
"""
from __future__ import print_function
import os
import sys
import ctypes
from ctypes import POINTER, pointer, c_int, c_long, c_char_p, c_double
import numpy as np

def load_feff8lpath():
    dllform = 'lib{:s}.so'
    pathsep = ':'
    loadlib = ctypes.cdll
    if sys.platform.lower().startswith('darwin'):
        dllform = 'lib{:s}.dylib'

    if os.name == 'nt':
        dllform = '{:s}.dll'
        pathsep = ';'
        loadlib = ctypes.windll

    dllname = dllform.format('feff8lpath')
    spath = ['.']
    spath.extend(os.environ.get('LD_LIBRARY_PATH','').split(pathsep))
    spath.extend(os.environ.get('PATH', '').split(pathsep))
    spath.extend(['../../local_install/lib/', '../../src/GENFMT/lib'])
    for dname in spath:
        fullname = os.path.join(dname, dllname)
        if os.path.exists(fullname):
            return loadlib.LoadLibrary(fullname)
    return None


FLIB = load_feff8lpath()


FEFF_maxpts = 150  # nex
FEFF_maxpot = 11   # nphx
FEFF_maxleg = 9    # legtot
BOHR = 0.5291772490


# bytes/str conversion
str2bytes = bytes2str = str
if sys.version_info[0] == 3:
    def bytes2str(val):
        if isinstance(val, str):
            return val
        if isinstance(val, bytes):
            return str(val, 'utf-8')
        return str(val)

    def str2bytes(val):
        if isinstance(val, bytes):
            return val
        return bytes(val, 'utf-8')

def with_phase_file(fcn):
    """decorator to ensure that the wrapped function either
    has a non-None 'phase_file' argument or that that
    self.phase_file is not None
    """
    errmsg = "function '%s' needs a non-None phase_file"
    def wrapper(*args, **keywords):
        "needs phase_file"
        phase_file = keywords.get('phase_file', None)
        if phase_file is None:
            phase_file = getattr(args[0], 'phase_file', None)
            if phase_file is None:
                raise AttributeError(errmsg % fcn.__name__)
        else:
            setattr(args[0], 'phase_file', phase_file)
        # raise Warning(errmsg % fcn.__name__)
        return fcn(*args, **keywords)
    wrapper.__doc__ = fcn.__doc__
    wrapper.__name__ = fcn.__name__
    wrapper.__filename__ = fcn.__code__.co_filename
    wrapper.__dict__.update(fcn.__dict__)
    return wrapper

class ScatteringPath(object):
    """A Scatering Path for calculating a XAFS signal with Feff

    A calculation requires a Potentials and Phase Shift calculation
    in PAD format from Feff85, and a list of scattering paths

    Usage:
    ------
       # create path
       path  = ScatteringPath(phase_file='phase.pad')

       # list 'ipot' and labels for absorber, scattererers
       path.list_scatterers()

       # set coords for absorbing atom
       path.set_absorber(x=0., y=0., z=0.)

       # add scattering atom
       path.add_scatterer(x=1.5, y=1.5, z=1.5, ipot=1)

       # calculate basic (unaltered) XAFS contributions
       path.calcuate_xafs()

    """
    def __init__(self, phase_file=None):
        self.phase_file = phase_file
        self.clear()

    def clear(self):
        """reset all path data"""

        self.index   = 1
        self.degen   = 1.
        self.nnnn_out = False
        self.json_out = False
        self.verbose  = False
        self.ipol   = 0
        self.ellip  = 0.
        self.nepts  = 0
        self.genfmt_order = 2
        self.genfmt_vers  = ""
        self.exch_label   = ""
        self.rs     = 0.
        self.vint   = 0.
        self.xmu    = 0.
        self.edge   = 0.
        self.kf     = 0.
        self.rnorman = 0.
        self.gamach = 0.
        self.nepts  = FEFF_maxpts

        dargs = dict(dtype=np.float64, order='F')
        largs = dict(dtype=np.int32, order='F')

        self.evec   = np.zeros(3, **dargs)
        self.xivec  = np.zeros(3, **dargs)
        self.ipot   = np.zeros(1+FEFF_maxleg, **largs)
        self.beta   = np.zeros(1+FEFF_maxleg, **dargs)
        self.eta    = np.zeros(2+FEFF_maxleg, **dargs)
        self.ri     = np.zeros(FEFF_maxleg, **dargs)
        self.rat    = np.zeros((3, 2+FEFF_maxleg), **dargs)
        self.iz     = np.zeros(1+FEFF_maxpot, **largs)
        self.kfeff  = np.zeros(FEFF_maxpts, **dargs)
        self.real_phc = np.zeros(FEFF_maxpts, **dargs)
        self.mag_feff = np.zeros(FEFF_maxpts, **dargs)
        self.pha_feff = np.zeros(FEFF_maxpts, **dargs)
        self.red_fact = np.zeros(FEFF_maxpts, **dargs)
        self.lam      = np.zeros(FEFF_maxpts, **dargs)
        self.rep      = np.zeros(FEFF_maxpts, **dargs)
        self.nleg = 1

    @with_phase_file
    def list_scatterers(self, phase_file=None):
        """list Feff Potentials atoms ('ipots') fo phase file"""
        atoms = []
        with open(self.phase_file,'r') as fh:
            line1_words = fh.readline().strip().split()
            text = fh.readlines()
        nphases = int(line1_words[4])
        for line in text[4:]:
            if line.startswith('$'): continue
            words = line.split()
            atoms.append((int(words[1]), words[2]))
            if len(atoms) > nphases:
                break
        out = ["# Potential   Z   Symbol"]
        for ipot, atom in enumerate(atoms):
            out.append("    %2i      %3i     %s" % (ipot, atom[0], atom[1]))
        return "\n".join(out)

    @with_phase_file
    def set_absorber(self, x=0., y=0., z=0., phase_file=None):
        """set coordinates for absorbing atom ('ipot'=0)"""
        self.rat[0, 0] = x
        self.rat[1, 0] = y
        self.rat[2, 0] = z

    @with_phase_file
    def add_scatterer(self, x=0., y=0., z=0., ipot=1, phase_file=None):
        self.rat[0, self.nleg] = x
        self.rat[1, self.nleg] = y
        self.rat[2, self.nleg] = z
        self.ipot[self.nleg] = ipot
        self.nleg += 1
        # set final atom coords to same as absorber
        self.rat[0, self.nleg] = self.rat[0, 0]
        self.rat[1, self.nleg] = self.rat[1, 0]
        self.rat[2, self.nleg] = self.rat[2, 0]
        self.ipot[self.nleg]   = self.ipot[0]

    @with_phase_file
    def calculate_xafs(self, phase_file=None):

        class args: pass

        # strings / char*.  Note fixed length to match Fortran
        args.phase_file     = (self.phase_file + ' '*256)[:256]
        args.exch_label     = str2bytes(' '*8)
        args.genfmt_version = str2bytes(' '*30)

        # integers, including booleans
        for attr in ('index', 'nleg', 'genfmt_order', 'ipol', 'nnnn_out',
                     'json_out', 'verbose', 'nepts'):
            setattr(args, attr, pointer(c_long(int(getattr(self, attr)))))

        # doubles
        for attr in ('degen', 'rs', 'vint', 'xmu', 'edge', 'kf', 'rnorman',
                     'gamach', 'ellip'):
            setattr(args, attr, pointer(c_double(getattr(self, attr))))

        # integer arrays
        for attr in ('ipot', 'iz'):
            arr = getattr(self, attr)
            cdata = arr.ctypes.data_as(POINTER(arr.size*c_int))
            setattr(args, attr, cdata)

        # double arrays
        self.rat = self.rat/BOHR
        for attr in ('evec', 'xivec', 'rat', 'ri', 'beta', 'eta',
                     'kfeff', 'real_phc', 'mag_feff', 'pha_feff',
                     'red_fact', 'lam', 'rep'):
            arr = getattr(self, attr)
            cdata = arr.ctypes.data_as(POINTER(arr.size*c_double))
            setattr(args, attr, cdata)

        x = FLIB.calc_onepath(args.phase_file, args.index, args.nleg,
                              args.degen, args.genfmt_order,
                              args.exch_label, args.rs, args.vint,
                              args.xmu, args.edge, args.kf, args.rnorman,
                              args.gamach, args.genfmt_version, args.ipot,
                              args.rat, args.iz, args.ipol, args.evec,
                              args.ellip, args.xivec, args.nnnn_out,
                              args.json_out, args.verbose, args.ri,
                              args.beta, args.eta, args.nepts, args.kfeff,
                              args.real_phc, args.mag_feff, args.pha_feff,
                              args.red_fact, args.lam, args.rep)

        self.exch_label = bytes2str(args.exch_label).strip()
        self.genfmt_version = bytes2str(args.genfmt_version).strip()

        for attr in ('index', 'nleg', 'genfmt_order', 'degen', 'rs',
                     'vint', 'xmu', 'edge', 'kf', 'rnorman', 'gamach',
                     'ipol', 'ellip', 'nnnn_out', 'json_out', 'verbose',
                     'nepts'):
            setattr(self, attr, getattr(args, attr).contents.value)

        for attr in ('ipot', 'evec', 'xivec', 'beta', 'eta', 'ri', 'rat',
                    'iz', 'kfeff', 'real_phc', 'mag_feff', 'pha_feff',
                    'red_fact', 'lam', 'rep'):
            setattr(self, attr, np.array(getattr(args, attr).contents[:]))

        # some data needs recasting, reformatting
        self.nnnn_out = bool(self.nnnn_out)
        self.json_out = bool(self.json_out)
        self.verbose  = bool(self.verbose)
        self.rat = self.rat.reshape((2+FEFF_maxleg, 3)).transpose()*BOHR


if __name__ == '__main__':
    path = ScatteringPath(phase_file='phase.pad')
    path.set_absorber( x=0.01,   y=0.1,   z=0.01)
    path.add_scatterer(x=1.8058, y=0.005, z=1.8063, ipot=1)
    path.degen = 12
    path.calculate_xafs()
    print('# Calculate EXAFS with PhaseFile: {:s}'.format(path.phase_file))
    print('# Path Geometry: \n#  IPOT  IZ     X        Y        Z')
    for i in range(path.nleg):
        ipot = path.ipot[i]
        iz   = path.iz[ipot]
        rat  = path.rat[:,i]
        print("#   %2i   %2i  %8.4f %8.4f %8.4f" % (ipot,iz, rat[0], rat[1], rat[2]))

    print("# Polarization: {:d}, ellipticity={:4f}".format(path.ipol, path.ellip))
    print("# Polarization E Vector = {:s}".format(", ".join(["%.4f" % a for a in path.evec])))
    print("# Polarization X Vector = {:s}".format(", ".join(["%.4f" % a for a in path.xivec])))
    print("# Path Settings")
    for attr in ('rs', 'vint', 'xmu', 'edge', 'kf', 'rnorman', 'gamach'):
          print("#   {:8s} = {:+4f} ".format(attr, getattr(path, attr)))
    for attr in ('exch_label', 'genfmt_version'):
          print("#   {:8s} = {:s} ".format(attr, getattr(path, attr)))

    print("Path settings:  degen=%10.5f,  xmu=%10.5f, kf=%10.5f" % (path.degen, path.xmu, path.kf))
    npts = 1 + max(np.where(path.kfeff > 0)[0])
    print("# k         rep          real_phc     phase_feff   mag_feff     red_factor   lambda ")
    fmt = " %6.3f  %11.7f  %11.7f  %11.7f  %11.7f  %11.7f  %11.7f"
    for i in range(int(npts/3.0)):
        print(fmt % (path.kfeff[i], path.rep[i], path.real_phc[i], path.pha_feff[i],
              path.mag_feff[i], path.red_fact[i], path.lam[i]))
        # print(fmt.format(path.kfeff[i], path.rep[i], path.real_phc[i],
        #                  path.mag_feff[i], path.red_fact[i], path.lam[i]))



## feff85exafs unit testing system using larch
## see HEADERS/license.h for feff's license information

import os
import sys
from os.path import isfile, isdir, join, exists
import re

from math import isnan

import nose
import pytest

import larch
from larch_plugins.xafs.feffdat import feffpath
from larch_plugins.io import read_ascii

from f85ut import ut

ALL_FOLDERS = ('Copper', 'NiO', 'UO2', 'Zircon', 'ferrocene',
               'bromoadamantane', 'LCO-para', 'LCO-perp')

class Feff8Test:
    def __init__(self, folder, verbose=True, doplot=False, doscf=None):
        if folder not in ALL_FOLDERS:
            raise ValueError("Must choose one of the available tests:\n  %s\n" % repr(ALL_FOLDERS))

        if doscf is None:
            envar = os.getenv('FEFF_TEST_SCF', 'False')
            doscf = envar.lower() in ("yes", "true", "on", "y", "t", "1")

        self.folder = folder
        self.test = ut(folder)
        self.test.verbose = verbose
        self.test.doplot = doplot
        self.test.doscf = doscf

    def test_feffrun(self):
        "run feff"
        self.test.run()
        scf = 'without SCF'
        if self.test.doscf:
            scf = 'with SCF'
        # feff8l.log shouldn't be more than a couple thousand lines long
        # (ferrocene w/SCF is 1096)
        with open(join(self.test.testrun, 'feff8l.log'), 'r') as log:
            lines = log.readlines()
            endlines = '_'.join(lines[-5:])
            m = 'Done with module 6:' in endlines

        assert m and self.test.available(1), "feff run on %s (%s) not successful" % (self.folder, scf)


    def test_feff(self):
        """check F_eff for each path"""
        assert self.test.feffran, "no test results found for %s" % self.folder
        for path in self.test.paths:
            index = int(path[4:8])
            for part in ('feff', 'amp', 'phase',
                         'lam', 'phc', 'red_fact', 'rep'):
                result = self.test.compare(index, part=part)
                assert result, "comparison of %s for path %d in %s" % (part, index, self.folder)

    def test_terms(self):
        "check various quantities computed by feff"
        assert self.test.feffran, "no test results found for %s" % self.folder
        for npath in range(1, 10):
            basefile = join(self.test.baseline, 'feff%4.4i.dat' % npath)
            if not exists(basefile):
                continue
            bl = feffpath(join(self.test.baseline, 'feff%4.4i.dat' % npath), _larch=self.test._larch)
            tr = feffpath(join(self.test.testrun,  'feff%4.4i.dat' % npath), _larch=self.test._larch)
            for term in ('edge', 'gam_ch', 'kf', 'mu', 'rs_int', 'vint'):
                blval = getattr(bl._feffdat, term)
                trval = getattr(tr._feffdat, term)
                tdiff = blval - trval
                assert abs(tdiff) < 2.0e-4, "feff term %s not close enough for %s" % (term, self.folder)

        for radius in ('muffintin', 'norman'):
            bl = self.test.radii('baseline', radius)
            tr = self.test.radii('testrun',  radius)
            assert bl == tr, "list of %s radii are different for %s" % (radius, self.folder)
        assert self.test.s02() == self.test.s02('baseline'), "s02 calculated incorrectly for %s" % self.folder

    def test_clean(self):
        """remove the test run"""
        self.test.clean()
        assert not isdir(self.test.testrun), "clean up %s calculation not successful" % self.folder

    def test_columns_wrapper(self):
        assert self.test.feffran, "no test results found for %s" % self.folder
        msg = "comparison of %s for path %i in %s using wrapper"

        for path in self.test.paths:
            index = int(path[4:8])
            for part in ('feff', 'amp', 'phase', 'lam', 'phc', 'red_fact', 'rep'):
                result = self.test.compare(index, part=part, use_wrapper=True)
                assert result, msg % (part, index, self.folder)


    def test_fit(self):
        """check various fitting and statistical parameters from a canned fit"""
        assert self.test.feffran, "no test results found for %s" % self.folder

        test_script = join(self.test.path, self.test.folder+'.py')
        stat_msg = "statistic %s evaluated inconsistently for %s (%.5f %.5f)"
        param_msg = "%s of fitting parameter %s evaluated inconsistently for %s (%.5f %.5f)"

        if isfile(test_script):
            self.test.fit()
            eps = self.test.epsfit
            for stat in ('chi_reduced', 'chi_square', 'rfactor'):
                bl = getattr(self.test.blfit, stat)
                tr = getattr(self.test.trfit, stat)
                close = abs((bl-tr)/bl) < eps
                assert close, stat_msg % (stat, self.folder, bl, tr)

            for par in self.test.blfit.var_names:
                for part, eps in (('value',  self.test.epsfit),
                                  ('stderr', self.test.epserr)):
                    bl = getattr(self.test.blfit.params[par], part)
                    tr = getattr(self.test.trfit.params[par], part)
                    close = abs((bl-tr)/bl) < eps
                    assert close, param_msg % (part, par, self.folder, bl, tr)


    @pytest.mark.skip(reason="not testing opconsat")
    def test_feffrun_opconsat(self):
        oca_ok = self.test.run_opconsat()
        assert oca_ok, "opconsat run on %s not successful" % folder

    @pytest.mark.skip(reason="not testing opconsat")
    def test_opconsat(self):
        check_skip("not checking opconstat")
        self.test.run_opconsat()
        for f in self.tests:
            pass # check_opconsat(f)


################################################################################


def check_opconsat(folder):
    orig = read_ascii(join(tests[folder].testrun, "..", "opconsat", "baseline", "exc.dat"), labels='a b c d', _larch=tests[folder]._larch)
    if not isfile(join(tests[folder].testrun, "exc.dat")):
        assert False, 'Failed to run opconsat'
    new  = read_ascii(join(tests[folder].testrun, "exc.dat"), labels='a b c d', _larch=tests[folder]._larch)
    #show(orig, _larch=self._larch)
    #show(new, _larch=self._larch)
    rf1  = sum((orig.a - new.a)**2) / sum(orig.a**2)
    rf2  = sum((orig.b - new.b)**2) / sum(orig.b**2)
    rf3  = sum((orig.c - new.c)**2) / sum(orig.c**2)
    rf4  = sum((orig.d - new.d)**2) / sum(orig.d**2)
    which = []
    failed = False
    if isnan(rf1):
        which.append('1')
        failed = True
    if isnan(rf2):
        which.append('2')
        failed = True
    if isnan(rf3):
        which.append('3')
        failed = True
    if isnan(rf4):
        which.append('4')
        failed = True
    if failed:
        assert False, "columns of exc.dat are not a number: %s" % ', '.join(which)

    which = []
    failed = False
    if rf1 < tests[folder].epsfit:
        which.append('1')
        failed = True
    if rf2 < tests[folder].epsfit:
        which.append('2')
        failed = True
    if rf3 < tests[folder].epsfit:
        which.append('3')
        failed = True
    if rf4 < tests[folder].epsfit:
        which.append('4')
        failed = True
    assert failed, "columns of exc.dat are not equal to baseline: %s" % ', '.join(which)

if __name__ == '__main__':
    TEST_FOLDERS = ('Copper', 'NiO', 'Zircon', 'ferrocene', 'LCO-para', 'LCO-perp')
    # TEST_FOLDERS = ALL_FOLDERS
    for folder in TEST_FOLDERS:
        t =  Feff8Test(folder)
        t.test_feffrun()
        t.test_feff()
        t.test_terms()
        t.test_columns_wrapper()
        t.test_fit()
        t.test_clean()



#!/usr/bin/python

import sys, subprocess, glob, re
from   os      import unlink, makedirs, chdir
from   os.path import isdir,  realpath, join
from   shutil  import rmtree
import pystache, json

from optparse import OptionParser
parser = OptionParser()
parser.add_option("-f", "--folder", dest="folder",
                  help="perform test for FOLDER", metavar="FOLDER")

parser.add_option("-s", "--scf",
                  action="store_true", dest="doscf", default=False,
                  help="perform Feff calculation with SCF (default is no SCF)")


(options, args) = parser.parse_args()

scf = 'noSCF'
if options.doscf: scf = 'withSCF'

repotop = realpath(join('..','..'))
if options.folder[-1] == '/': options.folder = options.folder[:-1]

## write the feff.inp file
target = join(options.folder, 'baseline', scf)

mat_json = json.load(open(join(options.folder, options.folder + '.json')))

mat_json['doscf']='* '
if options.doscf: mat_json['doscf']=''

if isdir(target): rmtree(target)
makedirs(target)

renderer = pystache.Renderer()
with open(join(target,'feff.inp'), 'w') as inp:
    inp.write(renderer.render_path( join(options.folder, options.folder + '.mustache'), # material/material.mustache
                                    mat_json ))                                         # material/material.json

chdir(target)


## location of script for running a version of feff85exafs from the
## beginning of unit test developments,
## https://github.com/xraypy/feff85exafs/commit/cac0f8c90749ce52581a658c5a6c8ae144cc2211
f85escript = join(repotop, 'bin', 'f85e')
## run the f85e shell script, which emulates the behavior of the monolithic Feff application
subprocess.call(f85escript);

## cull the files we don't need for testing
feffoutput = glob.glob("*")
for f in sorted(feffoutput):
    tosave = re.compile("feff(\d+\.dat|\.inp)|(chi|files|paths|xmu)\.dat|f85e.log")
    if not tosave.match(f): unlink(f)
